#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fftw3.h>
#include "interpd.h"
#include "utils.h"



int gen_xi_star(interplan *plan, double **xis1, double **xis2) {
	/* Generate xi star vectors */
	double *xistar1, *xistar2;
	double a, b, c, d;
	int i;
	
	if (plan == NULL)
		return -1;
	if (plan->degree < 1)
		return -1;
	if (xis1 == NULL || xis2 == NULL)
		return -1;
	
	/* define xi*1 and xi*2 according to the points family */
	if (plan->family == 1 || plan->family == 3) {
		*xis1 = (double *)malloc(sizeof(double)*(plan->degree+2));
		*xis2 = (double *)malloc(sizeof(double)*(plan->degree+3));
		xistar1 = *xis1;
		xistar2 = *xis2;
		a = plan->xMin;
		b = plan->xMax;
		c = plan->yMin;
		d = plan->yMax;
	} else if (plan->family == 2 || plan->family == 4) {
		*xis2 = (double *)malloc(sizeof(double)*(plan->degree+1));
		*xis1 = (double *)malloc(sizeof(double)*(plan->degree+2));
		xistar1 = *xis2;
		xistar2 = *xis1;
		a = plan->yMin;
		b = plan->yMax;
		c = plan->xMin;
		d = plan->xMax;
	} else
		return -1;
	
	/* I inverted the definitions of xistar1 and xistar2, so that xistar2
	 * is always the longest array. In this way, I can always use the following
	 * for loops to populate the arrays. */
	
	for (i = 0; i <= plan->degree; i++) {
		xistar1[i] = PAD2DOM(cos(PI*i/plan->degree), a, b);
		xistar2[i] = PAD2DOM(cos(PI*i/(plan->degree+1)), c, d);
	}
	
	xistar2[i] = PAD2DOM(cos(PI*i/(plan->degree+1)), c, d);
	
	return 0;
}

int gen_fihw(interplan *plan) {
	double *xis1, *xis2;
	double wstar;
	int i, j;
	int size = (plan->degree+1) * (plan->degree+2);
	
	if (gen_xi_star(plan, &xis1, &xis2) < 0)
		return -1;

	plan->fihw = (double *)fftw_malloc(size*sizeof(double));
	
	/* Set to zero the whole matrix vector. Blas might be faster than a for loop,
	 * because it may use a machine dependent SIMD instruction set. */
	DSCAL(size, 0.0, plan->fihw, 1);
	
	switch (plan->family) {
		case 1:
		for (i = 0; i <= plan->degree; i++)
			for (j = 0; j <= plan->degree+1; j++)
				if (!same_parity(i, j))
					set_element(plan->fihw, i, j, plan->degree+2, plan->funct(xis1[i],xis2[j]));
		break;
		
		case 2:
		for (i = 0; i <= plan->degree+1; i++)
			for (j = 0; j <= plan->degree; j++)
				if (!same_parity(i, j))
					set_element(plan->fihw, i, j, plan->degree+1, plan->funct(xis1[i],xis2[j]));
		break;
		
		case 3:
		for (i = 0; i <= plan->degree; i++)
			for (j = 0; j <= plan->degree+1; j++)
				if (same_parity(i, j))
					set_element(plan->fihw, i, j, plan->degree+2, plan->funct(xis1[i],xis2[j]));
		break;
		
		case 4:
		for (i = 0; i <= plan->degree+1; i++)
			for (j = 0; j <= plan->degree; j++)
				if (same_parity(i, j))
					set_element(plan->fihw, i, j, plan->degree+1, plan->funct(xis1[i],xis2[j]));
		break;
		
		default: // can't be, unless the memory is broken
		break;
	}
	
	// free the memory.
	free(xis1);
	free(xis2);
	
	/* It is possible to save some time multiplying by wstar just after computing
	 * the function. Moreover, if wstar = 2.0*wstar, we will save a further
	 * matrix-by-scalar multiplication later. */
	wstar = 1.0/(plan->degree*(plan->degree+1));
	DSCAL(size, wstar, plan->fihw, 1);
	
	return 0;
}

int gen_coeff(interplan *plan) {
	fftw_plan f_plan;
	int nx, ny;
	int i, j;
	
	// Prepare the sizes of the matrix.
	if (plan->family == 1 || plan->family == 3) {
		plan->nr_coeff = nx = plan->degree+1; // size of columns
		plan->nc_coeff = ny = plan->degree+2; // size of rows
	} else if (plan->family == 2 || plan->family == 4) {
		plan->nr_coeff = nx = plan->degree+2; // size of columns
		plan->nc_coeff = ny = plan->degree+1; // size of rows
	} else
		return -1;
	
	/* This matrix will have an exceeding row or column, compared to what
	 * is actually necessary. This will later set to 0. */
	plan->coeff = (double *)fftw_malloc(sizeof(double)*nx*ny);
	
	f_plan = fftw_plan_r2r_2d(nx, ny, plan->fihw, plan->coeff, FFTW_REDFT00, FFTW_REDFT00, plan->fftw_flags);
	
	fftw_execute(f_plan);
	fftw_destroy_plan(f_plan);
	
	/* I keep only the upper-left triangular. */
	for (i = 1; i < nx; i++)
		for (j = ny-i-(plan->family%2); j < ny; j++)
			set_element(plan->coeff,i,j,ny,0.0);
	
	// Set to 0 the exceeding column/row.
	if (nx > ny)
		DSCAL(ny, 0.0, plan->coeff+ny*(nx-1), 1);
	else
		DSCAL(nx, 0.0, plan->coeff+ny-1, ny);
	
	/* Final adjustments: forst I have to divide by two the element
	 * 0,0 of the matrix. */ 
	plan->coeff[0] = plan->coeff[0] / 2.0;
	
	/* Now I have to divide by SQRT2 the first row (but the first element,
	 * which I already did). */
	DSCAL(ny-1, INVSQRT2, plan->coeff+1, 1);
	
	/* and now the first column of the matrix (always but the first element). */
	DSCAL(nx-1, INVSQRT2, plan->coeff+ny, ny);
	
	/* I don't divide the element 0,0 twice by SQRT2 because I think it's
	 * more precise if I divide it only once by 2.0. */	
			
	/* Now I have b_{j,k-j}. I have to find the c_{j,k-j}. */
	
	i = (2 - (2 - plan->family%2)) * plan->degree;
	j = ((2 - plan->family%2) - 1) * plan->degree;
	set_element(plan->coeff,i,j,ny, 0.5*get_element(plan->coeff,i,j,ny)); 
	
	return 0;
}

int est_error(interplan *plan) {
	int j, k;
	
	if (plan == NULL)
		return -1;
	if (plan->coeff == NULL)
		return -1;
	
	plan->err_est = 0.0;
	
	for (k = plan->degree-2; k <= plan->degree; k++) {
		for (j = 0; j <= k; j++)
			plan->err_est += fabs(get_element(plan->coeff, j, k-j, plan->nc_coeff));
	}
	
	plan->err_est *= 2.0;
	
	return 0;
}
