#include <stdlib.h>
#include <stdio.h>
#include <fftw3.h>
#ifdef NORMAL_BLAS
#include <cblas.h>
#elif ACML_BLAS
#include <acml.h>
#endif
#include <nfft/nfft3.h>
#include <math.h>
#include "interpd.h"
#include "utils.h"

int load_fftw_wisdom() {
	FILE *wisdom;
	
	wisdom = fopen("pad_wisdom", "r");
	
	if (wisdom != NULL) {
		fftw_import_wisdom_from_file(wisdom);
		fclose(wisdom);
		
		return 0;
	}
	
	return -1;
}

int init_plan(interplan *p, int family, int degree, double (*funct)(double,double), double xMin, double xMax, double yMin, double yMax, unsigned fftw_flags) {
	if (family < 1 || family > 4 || degree < 1)
		return -1;
	
	if (p == NULL || funct == NULL)
		return -1;
	if (xMax <= xMin || yMax <= yMin)
		return -1;
		
	p->family = family;
	p->degree = degree;
	p->funct = funct;
	p->xMin = xMin;
	p->xMax = xMax;
	p->yMin = yMin;
	p->yMax = yMax;
		
	if (fftw_flags == 0)
		p->fftw_flags = FFTW_ESTIMATE;
	else
		p->fftw_flags = fftw_flags;		
	
	if (gen_fihw(p) < 0)
		return -1;
		
	if (gen_coeff(p) < 0)
		return -1;
	
	return 0;
}

int interpolate(interplan *p, int ntarget, double *target, int m, unsigned flags) {
	if (p == NULL || ntarget < 1 || target == NULL)
		return -1;
	
	/* Target must be an array twice as big as ntarget; it means that
	 * each target point is composed by the x and y coordinates. */
	p->ntarget = ntarget;
	p->target = target;
	
	if (m > 0)
		p->m = m;
	else
		p->m = 0;
	
	return func_on_targets(p, flags);
}

int destroy_plan(interplan *p) {
	if (p == NULL)
		return -1;
	
	/* Do not destroy the target points vector. We don't know how
	 * it's been allocated. */
	
	fftw_free(p->coeff);
	fftw_free(p->fihw);
	fftw_free(p->funct_on_target);
	
	return 0;
}

// Chebyshev polynomials of first kind.
int cheb0(int n, double pt, double *mat) {
	int i;
	
	if (mat == NULL)
		return -1;
	
	mat[0] = 1.0;
	mat[1] = SQRT2*pt;
	
	if (n <= 1)
		return 0;
	
	mat[2] = 2*pt*mat[1] - SQRT2*mat[0];
	
	if (n <= 2)
		return 0;
	
	/* There's no need any more to multiply by SQRT2, because the
	 * the polynomials of each degree will be already scaled. */ 
	for (i = 3; i <= n; i++)
		mat[i] = 2*pt*mat[i-1] - mat[i-2];
	
	return 0;
}

// First derivative of Chebyshev polynomials of first kind.
int cheb1(int n, double pt, double *mat) {
	int i;
	
	if (mat == NULL)
		return -1;
	
	mat[0] = 0.0;
	mat[1] = SQRT2;
	
	if (n <= 1)
		return 0;
		
	mat[2] = SQRT2*2*pt;
	
	if (n <= 2) {
		mat[2] *= 2;
		return 0;
	}
	
	// Same as before, as for the SQRT2.
	mat[3] = 2*pt*mat[2] - mat[1];
	
	if (n <= 3) {
		mat[2] *= 2;
		mat[3] *= 3;
		return 0;
	}
		
	// I have to stop at n-1, because this is the first deriv. 
	for (i = 4; i <= n; i++)
		mat[i] = 2*pt*mat[i-1] - mat[i-2];
	
	for (i = 0; i <= n; i++)
		mat[i] *= i;
	
	return 0;
}

int cheb2(int n, double pt, double *Tsec) {
	double *Tprime;
	int i;
	
	if (Tsec == NULL)
		return -1;
	
	Tprime = (double *)fftw_malloc(sizeof(double)*(n+1));
	cheb1(n, pt, Tprime);
	
	Tsec[0] = 0.0;
	Tsec[1] = 0.0;
	
	if (n <= 1)
		return 0;
	
	/* Here we don't need to multiply by SQRT2 at all, because the
	 * Tprime are already scaled. */
	Tsec[2] = 4*Tprime[1];
	
	if (n <= 2)
		return 0;
	
	for (i = 3; i <= n; i++)
		Tsec[i] = 2*pt*Tsec[i-1] + 4*Tprime[i-1] - Tsec[i-2];
	
	fftw_free(Tprime);
	
	return 0;
}

int interpolate_bl(interplan *p, int ntarget, double *target) {
	if (p == NULL || ntarget < 1 || target == NULL)
		return -1;
	
	/* Target must be an array twice as big as ntarget; it means that
	 * each target point is composed by the x and y coordinates. */
	p->ntarget = ntarget;
	p->target = target;
	
	return bl_func_on_targets(p, cheb0, cheb0);
}

int interpolate_d1(interplan *p, int ntarget, double *target, int var) {
	int ires = 0;
	double scalfact;
	
	if (p == NULL || ntarget < 1 || target == NULL)
		return -1;
	
	/* Target must be an array twice as big as ntarget; it means that
	 * each target point is composed by the x and y coordinates. */
	p->ntarget = ntarget;
	p->target = target;
	
	if (var == 0) {
		ires = bl_func_on_targets(p, cheb1, cheb0);
		scalfact = 2.0 / (p->xMax-p->xMin);
	} else if (var == 1) {
		ires = bl_func_on_targets(p, cheb0, cheb1);
		scalfact = 2.0 / (p->yMax-p->yMin);
	} else
		return -1;
	
	DSCAL(p->ntarget, scalfact, p->funct_on_target, 1);
	return ires;
}

int interpolate_d2(interplan *p, int ntarget, double *target, int var) {
	int ires = 0;
	double scalfact;
	
	if (p == NULL || ntarget < 1 || target == NULL)
		return -1;
	
	/* Target must be an array twice as big as ntarget; it means that
	 * each target point is composed by the x and y coordinates. */
	p->ntarget = ntarget;
	p->target = target;
	
	if (var == 0) {
		ires = bl_func_on_targets(p, cheb2, cheb0);
		scalfact = 4.0 / pow((p->xMax-p->xMin),2);;
	} else if (var == 1) {
		ires = bl_func_on_targets(p, cheb0, cheb2);
		scalfact = 4.0 / pow((p->yMax-p->yMin),2);
	} else if (var == 2) {
		ires = bl_func_on_targets(p, cheb1, cheb1);
		scalfact = 4.0 / ((p->xMax-p->xMin) * (p->yMax-p->yMin));
	} else
		return -1;
	
	DSCAL(p->ntarget, scalfact, p->funct_on_target, 1);
	return ires;
}
