/* For the coefficients:
 * 1. Generate the superset of Padua points, in matrix format.
 * 2. Set to 0 all the elements of the matrix.
 * 3. Set the right elements of the matrix to the value of the function.
 * 4. Multiply by 2 the matrix with BLAS (just a for loop, but maybe something
 * better in other architectures).
 * 5. Divide by SQRT2 the first row and the diagonal of the matrix.
 */

#ifndef INTERPD_H_
#define INTERPD_H_

#include <nfft/nfft3.h>
#include <nfft/util.h>
#include BLASINC

#define INVDOUBLEPI 0.1591549430918953 // 1/(2*PI)
#define SQRT2 1.4142135623730951
#define INVSQRT2 0.7071067811865476 // 1/sqrt(2)

#define EXACT 0x1
#define APPROX 0x0

/* Macros to get and set elements in a vector that represents a
 * matrix in row-major format.
 * matrix: pointer to vector
 * i: row number
 * j: colum number
 * n: number of columns (size of rows)
 * el: the element to be inserted
 */
#define get_element(matrix, i, j, n) (matrix[j+(i)*(n)])
#define set_element(matrix, i, j, n, el) (matrix[j+(i)*(n)] = el)

/* macro to verify if a and be have the same parity */
#define same_parity(a,b) ((a % 2 == 0 && b % 2 == 0) || (a % 2 != 0 && b % 2 != 0))

/* Define BLAS function names. */
#define append(a,b) a##b
#define aappend(a,b) append(a,b)
#define DSCAL aappend(BLASPREF,dscal)
#define DCOPY aappend(BLASPREF,dcopy)
#define DDOT aappend(BLASPREF,ddot)

typedef struct {
	int family;
	int degree;
	double *fihw;
	double *coeff;
	double xMin, xMax;
	double yMin, yMax;
	int nr_coeff; // number of rows (size of columns)
	int nc_coeff; // number of columns (size of rows)
	int ntarget; // num of target points (length of target/2)
	double *target;
	double (*funct)(double, double);
	double *funct_on_target;
	unsigned fftw_flags;
	int m;
	double err_est;
} interplan;


int gen_xi_star(interplan *plan, double **xis1, double **xis2);
int gen_fihw(interplan *);
int gen_coeff(interplan *);
int est_error(interplan *);
int func_on_targets(interplan *, unsigned);
int bl_func_on_targets(interplan *, int (*chebX)(int,double,double*), int (*chebY)(int,double,double*));

int load_fftw_wisdom();
int init_plan(interplan *p, int family, int degree, double (*funct)(double,double), double xMin, double xMax, double yMin, double yMax, unsigned fftw_flags);
int interpolate(interplan *p, int ntarget, double *target, int m, unsigned flags);
int interpolate_bl(interplan *p, int ntarget, double *target);
int interpolate_d1(interplan *p, int ntarget, double *target, int var);
int interpolate_d2(interplan *p, int ntarget, double *target, int var);
int destroy_plan(interplan *p);

#endif /*INTERPD_H_*/
