#include <stdlib.h>
#include <fftw3.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/uio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include "utils.h"
#include "interpd.h"

int generate_square_grid(int n, double **buffer, double xMin, double xMax, double yMin, double yMax) {
	double *b;
	double step, val;
	int i, j;
	
	if (n < 2 || buffer == NULL)
		return -1;
	if (xMax <= xMin || yMax <= yMin)
		return -1;
	
	*buffer = (double *)fftw_malloc(sizeof(double)*n*n*2);
	b = *buffer;
	
	step = (xMax-xMin) / (n-1);
	
	for (i = 0; i < n; i++) {
		val = MAX(MIN(xMin + step*i, xMax), xMin);
		
		for (j = 0; j < n; j++)
			set_element(b, j+i*n, 0, 2, val);
	}
	
	step = (yMax-yMin) / (n-1);
	
	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			val = MAX(MIN(yMin + step*j, yMax), yMin);
			set_element(b, j+i*n, 1, 2, val);
		}
	}
	
	return 0;	
}

int generate_rect_grid(int nx, int ny, double **buffer, double xMin, double xMax, double yMin, double yMax) {
	double *b;
	double step, val;
	int i, j;
	
	if (nx < 2 || ny < 2 || buffer == NULL)
		return -1;
	if (xMax <= xMin || yMax <= yMin)
		return -1;
	
	*buffer = (double *)fftw_malloc(sizeof(double)*nx*ny*2);
	b = *buffer;
	
	step = (xMax-xMin) / (nx-1);
	
	for (i = 0; i < nx; i++) {
		val = MAX(MIN(xMin + step*i, xMax), xMin);
		
		for (j = 0; j < ny; j++)
			set_element(b, j+i*nx, 0, 2, val);
	}
	
	step = (yMax-yMin) / (ny-1);
	
	for (i = 0; i < nx; i++) {
		for (j = 0; j < ny; j++) {
			val = MAX(MIN(yMin + step*j, yMax), yMin);
			set_element(b, j+i*nx, 1, 2, val);
		}
	}
	
	return 0;	
}

int generate_random(int n, double **buffer, double xMin, double xMax, double yMin, double yMax) {
	double *b;
	int i;
	
	if (n < 2 || buffer == NULL)
		return -1;
	if (xMax <= xMin || yMax <= yMin)
		return -1;
		
	srand48(time(NULL));
	
	*buffer = (double *)fftw_malloc(sizeof(double)*n*2);
	b = *buffer;
	
	/* Uniform random numbers in [xMin,xMax]x[yMin,yMax] */
	for (i = 0; i < n; i++) {
		b[2*i] = MAX(MIN((xMax-xMin) * drand48() + xMin, xMax), xMin);
		b[2*i+1] = MAX(MIN((yMax-yMin) * drand48() + yMin, yMax), yMin);
	}
	
	return 0;
}

int generate_padua_points(int n, double **buffer, int fam, double xMin, double xMax, double yMin, double yMax) {
	double *buf;
	double *xis1, *xis2;
	int i, j, k;
	int len_i, len_j;
	interplan plan;
	
	if (buffer == NULL)
		return -1;
	if (fam < 1 || fam > 4)
		return -1;
	if (n < 1)
		return -1;
	
	plan.degree = n;
	plan.family = fam;
	plan.xMin = xMin;
	plan.xMax = xMax;
	plan.yMin = yMin;
	plan.yMax = yMax;
	
	*buffer = (double *)fftw_malloc(sizeof(double)*(n+1)*(n+2));
	buf = *buffer;
	
	gen_xi_star(&plan, &xis1, &xis2);
	
	if (fam == 1 || fam == 3) {
		len_i = n;
		len_j = n+1;
	} else {
		len_i = n+1;
		len_j = n;
	}
	
	k = 0;
	for (i = 0; i <= len_i; i++) {
		for (j = 0; j <= len_j; j++) {
			if ((fam == 1 || fam == 2) && !same_parity(i,j)) {
				buf[2*k] = xis1[i];
				buf[2*k+1] = xis2[j];
				k++;
			} else if ((fam == 3 || fam == 4) && same_parity(i,j)) {
				buf[2*k] = xis1[i];
				buf[2*k+1] = xis2[j];
				k++;
			}
		}
	}
	
	fftw_free(xis1);
	fftw_free(xis2);
	
	return (n+1)*(n+2)/2;
}

/* Subtract the `struct timeval' values X and Y,
 * storing the result in RESULT.
 * Return 1 if the difference is negative, otherwise 0.  */
int timeval_subtract (result, x, y)
	struct timeval *result, *x, *y; {
	/* Perform the carry for the later subtraction by updating y. */
	if (x->tv_usec < y->tv_usec) {
		int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
		y->tv_usec -= 1000000 * nsec;
		y->tv_sec += nsec;
	}
	if (x->tv_usec - y->tv_usec > 1000000) {
		int nsec = (x->tv_usec - y->tv_usec) / 1000000;
		y->tv_usec += 1000000 * nsec;
		y->tv_sec -= nsec;
	}
     
	/* Compute the time remaining to wait.
	 * tv_usec is certainly positive. */
	result->tv_sec = x->tv_sec - y->tv_sec;
	result->tv_usec = x->tv_usec - y->tv_usec;
     
	/* Return 1 if result is negative. */
	return x->tv_sec < y->tv_sec;
}

int save_data(interplan *plan, char *filename) {
	FILE *out;
	int i;
	
	if (plan == NULL)
		return -1;
	
	if (filename == NULL)
		return -1;
	
	out = fopen(filename, "w");
	
	if (out == NULL)
		return -1;
	
	fprintf(out, "# Function evaluated at %d target points.\n", plan->ntarget);
	
	for (i = 0; i < plan->ntarget; i++)
		fprintf(out, "%.8E %.8E %.8E\n", plan->target[2*i], plan->target[2*i+1], plan->funct_on_target[i]);
	
	fprintf(out, "# End of data.\n");
	fclose(out);
	
	return 0;
}

int plot_data(interplan *plan, char *filename) {
	int i;
	char *file;
	FILE *data;
	char *script;
	char *plot_opts = "";
	int temp = 0;
	
	// System data.
	pid_t childpid;
	int fdesc[2];
	
	if (plan == NULL)
		return -1;
		
	if (filename == NULL) {
		file = "/tmp/interp_data.dat";
		temp = 1;
	} else
		file = filename;
	
	data = fopen(file, "w");
	
	if (data == NULL)
		return -1;
	
	fprintf(data, "# InterPD data file. Function evaluated at target points.\n");
	
	for (i = 0; i < plan->ntarget; i++)
		fprintf(data, "%.8E %.8E %.8E\n", plan->target[2*i], plan->target[2*i+1], plan->funct_on_target[i]);
	
	fclose(data);
	
	script = (char*)malloc(1024*sizeof(char));
	memset(script, 0, 1024*sizeof(char));
	
	if (plan->ntarget > 5000)
		plot_opts = " with dots";
	
	snprintf(script, 1024, "set grid xtics ytics ztics\n"
	"set title \'Function evaluated at target points.\'\n"
	"splot \'%s\'%s\n", file, plot_opts);

//	For debug purposes.	
//	printf("%s", script);
	
	if (pipe(fdesc) < 0) {
		perror("pipe failed");
		return -1;
	}
	
	childpid = fork();
	
	if (childpid < 0)
		perror("fork() failed");
	else if (childpid > 0) { // parent
		int status;
		
		// Close reading side of the pipe.
		if (close(fdesc[0]) < 0)
			perror("close() failed");
		
		// Send data to pipe.
		for (i = 0; i < 1024*sizeof(char); i++) {
			if (write(fdesc[1], &script[i], sizeof(char)) < 0) {
				perror("write() failed");
				return -1;
			}
			
			if (script[i] == 0)
				break;
		}
		
		// Close pipe.
		if (close(fdesc[1]) < 0)
			perror("close() failed");

		// Wait for the child to exit.
		wait(&status);
		if (WEXITSTATUS(status) == 53)
			perror("gnuplot not found");
		
		if (temp != 0) {
			if (unlink(file) < 0)
				perror("unlink() failed");
		}
	} else { // (childpid == 0) Now I'm in the child process
		// Set reading side of pipe to stdin.
		if (dup2(fdesc[0],0) < 0)
			perror("dup() failed");
		
		// Close writing side of pipe.
		if (close(fdesc[1]) < 0)
			perror("close() failed");
		
		if (execlp("gnuplot", "gnuplot", "-persist", NULL) < 0) {
			perror("exec() failed");
			close(fdesc[0]);
			exit(53);
		}
	}
	
	return 0;
}

double eval_error(interplan *plan) {
	double maxErr, maxDev;
	double val, fval, minVal, maxVal, meanVal;
	int i;
	
	maxErr = meanVal = 0;
	minVal = HUGE_VAL;
	maxVal = -HUGE_VAL;
	
	for (i = 0; i < plan->ntarget; i++) {
		fval = plan->funct(plan->target[2*i], plan->target[2*i+1]);
		val = fabs(plan->funct_on_target[i] - fval);
		
		if (val > maxErr)
			maxErr = val;	
			
		maxVal = MAX(fval, maxVal);
		minVal = MIN(fval, minVal);
		meanVal += fval;		
	}
		
	meanVal = meanVal / (double) plan->ntarget;
	maxDev = MAX(meanVal - minVal, maxVal - meanVal);
	
	return maxErr / maxDev;
}
